/**
 * Copyright (C) 2014 Felipe Ribeiro
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */
'use strict';
var Comparator = require('../util/comparator');

var COLOR_RED = 0;
var COLOR_BLACK = 1;

var NullNode = {
  color: COLOR_BLACK
};

function Node(value) {
  this.value = value;
  this.left = NullNode;
  this.right = NullNode;
  this.parent = NullNode;
  this.color = COLOR_RED;

  Object.defineProperties(this, {
    'grandparent': {
      get: function () {
        return this.parent && this.parent.parent;
      }
    },
    'uncle': {
      get: function () {
        var g = this.grandParent();
        return g && (this.parent === g.left ? g.right : g.left);
      }
    }
  });

}

/**
 * Red Black Tree
 *
 * Properties:
 * 1. A node is either red or black.
 * 2. The root is black.
 * 3. All leaves (NIL) are black.
 * 4. If a node is red, then both children are black.
 * 5. For each node, all paths from the node to descendant leaves contain
 * the same number of black nodes
 */
function RedBlackTree(compareFn) {
  this.root = null;
  this._comparator = new Comparator(compareFn);
}

RedBlackTree.prototype._bstInsert = function (node, parent) {
  if (!parent) {
    if (!this.root) {
      return;
    }
    parent = this.root;
  }
  var child = this._comparator.lessThan(node.value, parent.value) ?
    'left' : 'right';
  if (parent[child] !== NullNode) {
    this._bstInsert(node, parent[child]);
  } else {
    parent[child] = node;
    node.parent = parent;
  }
};

/**
 * In case 1 the tree is empty and the node is painted black
 * to satisfy property #2
 */
RedBlackTree.prototype._insertCase1 = function (node) {
  if (node.parent == NullNode)
    node.color = COLOR_BLACK;
  else
    this._insertCase2(node);
};

/**
 * In case 2 the node's parent is black, so property #4 is not invalidated
 */
RedBlackTree.prototype._insertCase2 = function (node) {
  if (node.parent.color == COLOR_BLACK)
    return; // Tree is still valid
  else
    this._insertCase3(node);
};

/**
 * In case 3, both the parent and the uncle are red, and need to be repainted
 * to satisfy property #5
 */
RedBlackTree.prototype._insertCase3 = function (node) {
  if (node.parent.color == COLOR_RED &&
      node.uncle != NullNode &&
      node.uncle.color == COLOR_RED) {

    node.parent.color = COLOR_BLACK;
    node.uncle.color = COLOR_BLACK;
    var g = node.grandparent;
    g.color = COLOR_RED;
    this._insertCase1(g);
  } else {
    this._insertCase4(node);
  }
};

/**
 * In case 4, the parent is red but the uncle is black;
 */
RedBlackTree.prototype._insertCase4 = function (node) {
  var g = node.grandparent;

  if (node == node.parent.right && node.parent == g.left) {
    this._rotateLeft(node.parent);
    node = node.left;
  } else if (node == node.parent.left && node.parent == g.right) {
    this._rotateRight(node.parent);
    node = node.right;
  }
  this._insertCase5(node);
};

RedBlackTree.prototype._insertCase5 = function (node) {
  node.parent.color = COLOR_BLACK;
  node.grandparent.color = COLOR_RED;
  if (node === node.parent.left)
    this._rotateRight(node.grandparent);
  else
    this._rotateRight(node.grandparent);
};

RedBlackTree.prototype._rotateLeft = function (node) {
  var rightChild = node.right;
  var parent = node.parent;
  var side = (parent.left === node ? 'left' : 'right');
  node.right = rightChild.left;
  rightChild.left = node;
  parent[side] = rightChild;

  rightChild.parent = parent;
  node.parent = rightChild;
};

RedBlackTree.prototype._rotateRight = function (node) {
  var leftChild = node.left;
  var parent = node.parent;
  var side = (parent.left === node ? 'left' : 'right');
  node.left = leftChild.right;
  leftChild.right = node;
  parent[side] = leftChild;

  leftChild.parent = parent;
  node.parent = leftChild;
};


RedBlackTree.prototype.insert = function (value) {
  var n = new Node(value);

  this._bstInsert(n);
  this._insertCase1(n);
};

module.exports = RedBlackTree;
